@@ -1,14 +1,15 @@
 //! Corrections that can be performed on visibility data
 
 use crate::{
-    ndarray::{parallel::prelude::*, Array3, Axis},
-    Jones,
+    ndarray::{parallel::prelude::*, Array2, Array3, Axis},
+    BirliError, Jones,
 };
 use indicatif::{ProgressBar, ProgressDrawTarget, ProgressStyle};
+use itertools::izip;
 use log::trace;
 use marlu::{
-    constants::VEL_C, mwalib::CorrelatorContext, precession::precess_time, time, Complex,
-    LatLngHeight, RADec, XyzGeodetic, UVW,
+    constants::VEL_C, hifitime::Epoch, mwalib::CorrelatorContext, precession::precess_time,
+    Complex, LatLngHeight, RADec, XyzGeodetic, UVW,
 };
 use std::{f64::consts::PI, ops::Range};
 
@@ -39,25 +40,25 @@ use std::{f64::consts::PI, ops::Range};
 /// let context = CorrelatorContext::new(&metafits_path, &gpufits_paths).unwrap();
 ///
 /// // Determine which timesteps and coarse channels we want to use
-/// let img_coarse_chan_idxs = &context.common_coarse_chan_indices;
-/// let img_timestep_idxs = &context.common_timestep_indices;
+/// let sel_coarse_chan_idxs = &context.common_coarse_chan_indices;
+/// let sel_timestep_idxs = &context.common_timestep_indices;
 /// let baseline_idxs = (0..context.metafits_context.num_baselines).collect::<Vec<_>>();
 ///
-/// let img_timestep_range =
-///     *img_timestep_idxs.first().unwrap()..(*img_timestep_idxs.last().unwrap() + 1);
-/// let img_coarse_chan_range =
-///     *img_coarse_chan_idxs.first().unwrap()..(*img_coarse_chan_idxs.last().unwrap() + 1);
+/// let sel_timestep_range =
+///     *sel_timestep_idxs.first().unwrap()..(*sel_timestep_idxs.last().unwrap() + 1);
+/// let sel_coarse_chan_range =
+///     *sel_coarse_chan_idxs.first().unwrap()..(*sel_coarse_chan_idxs.last().unwrap() + 1);
 ///
 /// // read visibilities out of the gpubox files
 /// let (mut jones_array, _) = context_to_jones_array(
 ///     &context,
-///     &img_timestep_range,
-///     &img_coarse_chan_range,
+///     &sel_timestep_range,
+///     &sel_coarse_chan_range,
 ///     None,
 ///     false,
 /// ).unwrap();
 ///
-/// correct_cable_lengths(&context, &mut jones_array, &img_coarse_chan_range, false);
+/// correct_cable_lengths(&context, &mut jones_array, &sel_coarse_chan_range, false);
 /// ```
 ///
 /// # Accuracy
@@ -106,7 +107,6 @@ pub fn correct_cable_lengths(
     );
     correction_progress.set_message("cable corrections");
 
-    // TODO: parallelize
     jones_array
         .axis_iter_mut(Axis(2))
         .into_par_iter()
@@ -119,7 +119,6 @@ pub fn correct_cable_lengths(
             let ant1 = &antennas[baseline.ant1_index];
             let ant2 = &antennas[baseline.ant2_index];
 
-            // TODO: I don't think this is technically in the right order
             let pol_lengths = [
                 ant2.rfinput_x.electrical_length_m - ant1.rfinput_x.electrical_length_m,
                 ant2.rfinput_y.electrical_length_m - ant1.rfinput_y.electrical_length_m,
@@ -179,20 +178,20 @@ pub fn correct_cable_lengths(
 /// let context = CorrelatorContext::new(&metafits_path, &gpufits_paths).unwrap();
 ///
 /// // Determine which timesteps and coarse channels we want to use
-/// let img_coarse_chan_idxs = &context.common_coarse_chan_indices;
-/// let img_timestep_idxs = &context.common_timestep_indices;
+/// let sel_coarse_chan_idxs = &context.common_coarse_chan_indices;
+/// let sel_timestep_idxs = &context.common_timestep_indices;
 /// let baseline_idxs = (0..context.metafits_context.num_baselines).collect::<Vec<_>>();
 ///
-/// let img_timestep_range =
-///     *img_timestep_idxs.first().unwrap()..(*img_timestep_idxs.last().unwrap() + 1);
-/// let img_coarse_chan_range =
-///     *img_coarse_chan_idxs.first().unwrap()..(*img_coarse_chan_idxs.last().unwrap() + 1);
+/// let sel_timestep_range =
+///     *sel_timestep_idxs.first().unwrap()..(*sel_timestep_idxs.last().unwrap() + 1);
+/// let sel_coarse_chan_range =
+///     *sel_coarse_chan_idxs.first().unwrap()..(*sel_coarse_chan_idxs.last().unwrap() + 1);
 ///
 /// // read visibilities out of the gpubox files
 /// let (mut jones_array, _) = context_to_jones_array(
 ///     &context,
-///     &img_timestep_range,
-///     &img_coarse_chan_range,
+///     &sel_timestep_range,
+///     &sel_coarse_chan_range,
 ///     None,
 ///     false,
 /// ).unwrap();
@@ -200,8 +199,8 @@ pub fn correct_cable_lengths(
 /// correct_geometry(
 ///     &context,
 ///     &mut jones_array,
-///     &img_timestep_range,
-///     &img_coarse_chan_range,
+///     &sel_timestep_range,
+///     &sel_coarse_chan_range,
 ///     None,
 ///     None,
 ///     false,
@@ -269,8 +268,9 @@ pub fn correct_geometry(
         .into_par_iter()
         .zip(timesteps)
         .for_each(|(mut jones_timestep_view, timestep)| {
-            let epoch =
-                time::gps_to_epoch(timestep.gps_time_ms as f64 / 1000.0 + integration_time_s / 2.0);
+            let epoch = Epoch::from_gpst_seconds(
+                timestep.gps_time_ms as f64 / 1000.0 + integration_time_s / 2.0,
+            );
 
             let prec_info = precess_time(
                 phase_centre_ra,
@@ -306,19 +306,166 @@ pub fn correct_geometry(
     trace!("end correct_geometry");
 }
 
+/// Apply corrections for digital gains for each coarse channel from values in metafits.
+///
+/// The channels provided in `jones_array` should correspond to the selected coarse channel range in
+/// `sel_coarse_chan_range`, such that `jones_array.dim().1 = num_fine_chans_per_coarse * sel_coarse_chan_range.len()`
+///
+/// # Arguments
+///
+/// - `context` - The correlator context.
+/// - `jones_array` - The array of Jones matrices to be corrected, [timestep][channel][baseleine].
+/// - `sel_coarse_chan_range` - The range of mwalib coarse channels which are used in in the channel
+///     dimension of the jones array.
+///
+/// # Assumptions
+/// - the digital gains are provided in [`mwalib::Rfinput.digital_gains`] in the same order as the
+///   coarse channel indices (increasing sky frequency)
+/// - the gains for the x rfinput are the same as the y rfinput for each antenna
+///
+/// # Errors
+/// - Will throw BadArrayShape if:
+///     - `jones_array.dim().1 != num_fine_chans_per_coarse * sel_coarse_chan_range.len()
+///     - `jones_array.dim().2 != sel_baselines.len()`
+pub fn correct_digital_gains(
+    context: &CorrelatorContext,
+    jones_array: &mut Array3<Jones<f32>>,
+    sel_coarse_chan_range: &Range<usize>,
+    // The tile index pairs for each selected baseline
+    sel_baselines: &[(usize, usize)],
+) -> Result<(), BirliError> {
+    let num_fine_chans_per_coarse = context.metafits_context.num_corr_fine_chans_per_coarse;
+
+    // TODO: proper error handling
+    let vis_dims = jones_array.dim();
+    if vis_dims.1 != sel_coarse_chan_range.len() * num_fine_chans_per_coarse {
+        return Err(BirliError::BadArrayShape {
+            argument: "sel_coarse_chan_range".into(),
+            function: "correct_digital_gains".into(),
+            expected: format!(
+                "(vis_dims.1={}) / (num_fine_chans_per_coarse={}) = {}",
+                vis_dims.1,
+                num_fine_chans_per_coarse,
+                vis_dims.1 / num_fine_chans_per_coarse
+            ),
+            received: format!("{:?}", sel_coarse_chan_range.len()),
+        });
+    }
+    if vis_dims.2 != sel_baselines.len() {
+        return Err(BirliError::BadArrayShape {
+            argument: "sel_baselines".into(),
+            function: "correct_digital_gains".into(),
+            expected: format!("vis_dims.2={}", vis_dims.2,),
+            received: format!("{:?}", sel_baselines.len()),
+        });
+    }
+    assert!(vis_dims.2 == sel_baselines.len());
+
+    let gains = Array2::from_shape_fn(
+        (
+            context.metafits_context.num_ants,
+            sel_coarse_chan_range.len(),
+        ),
+        |(ant_idx, coarse_chan_idx)| {
+            let ant = context.metafits_context.antennas[ant_idx].clone();
+            let mwalib_coarse_chan_idx = sel_coarse_chan_range.start + coarse_chan_idx;
+            // assert_eq!(
+            //     ant.rfinput_x.digital_gains[mwalib_coarse_chan_idx],
+            //     ant.rfinput_y.digital_gains[mwalib_coarse_chan_idx],
+            //     "x and y digital gains for antenna {:?} are not equal for coarse channel {}",
+            //     ant,
+            //     mwalib_coarse_chan_idx
+            // );
+            (
+                ant.rfinput_x.digital_gains[mwalib_coarse_chan_idx],
+                ant.rfinput_y.digital_gains[mwalib_coarse_chan_idx],
+            )
+        },
+    );
+
+    _correct_digital_gains(
+        jones_array,
+        &gains,
+        sel_baselines,
+        num_fine_chans_per_coarse,
+    )
+}
+
+fn _correct_digital_gains(
+    jones_array: &mut Array3<Jones<f32>>,
+    gains: &Array2<(f64, f64)>,
+    sel_baselines: &[(usize, usize)],
+    num_fine_chans_per_coarse: usize,
+) -> Result<(), BirliError> {
+    // TODO: proper error handling
+    let vis_dims = jones_array.dim();
+    let gain_dims = gains.dim();
+    if vis_dims.1 != gain_dims.1 * num_fine_chans_per_coarse {
+        return Err(BirliError::BadArrayShape {
+            argument: "gains".into(),
+            function: "_correct_digital_gains".into(),
+            expected: format!(
+                "(_, n, _), where n = (vis_dims.1={}) / (num_fine_chans_per_coarse={}) = {}",
+                vis_dims.1,
+                num_fine_chans_per_coarse,
+                vis_dims.1 / num_fine_chans_per_coarse
+            ),
+            received: format!("{:?}", gain_dims),
+        });
+    }
+    assert!(vis_dims.1 == gain_dims.1 * num_fine_chans_per_coarse);
+
+    // iterate through the selected baselines
+    for (mut jones_array, (ant1_idx, ant2_idx)) in izip!(
+        jones_array.axis_iter_mut(Axis(2)),
+        sel_baselines.iter().cloned()
+    ) {
+        // iterate through the selected coarse channels
+        for (mut jones_array, &(gain1x, gain1y), &(gain2x, gain2y)) in izip!(
+            jones_array.axis_chunks_iter_mut(Axis(1), num_fine_chans_per_coarse),
+            gains.index_axis(Axis(0), ant1_idx),
+            gains.index_axis(Axis(0), ant2_idx),
+        ) {
+            // for all visibilities in the selected coarse channel, for all timesteps
+            for jones in jones_array.iter_mut() {
+                // promote
+                let corrected = Jones::<f64>::from(*jones);
+                // divide by gain and demote
+                *jones = Jones::<f32>::from([
+                    (corrected[0].re / gain1x / gain2x) as _,
+                    (corrected[0].im / gain1x / gain2x) as _,
+                    (corrected[1].re / gain1x / gain2y) as _,
+                    (corrected[1].im / gain1x / gain2y) as _,
+                    (corrected[2].re / gain1y / gain2x) as _,
+                    (corrected[2].im / gain1y / gain2x) as _,
+                    (corrected[3].re / gain1y / gain2y) as _,
+                    (corrected[3].im / gain1y / gain2y) as _,
+                ]);
+            }
+        }
+    }
+
+    Ok(())
+}
+
 #[cfg(test)]
 mod tests {
     #![allow(clippy::float_cmp)]
 
-    use super::{correct_cable_lengths, correct_geometry, VEL_C};
+    use super::{correct_cable_lengths, correct_digital_gains, correct_geometry, VEL_C};
     use marlu::{
-        mwalib::CorrelatorContext, precession::precess_time, time, Complex, Jones, LatLngHeight,
-        RADec, XyzGeodetic, UVW,
+        hifitime::Epoch, mwalib::CorrelatorContext, precession::precess_time, Complex, Jones,
+        LatLngHeight, RADec, XyzGeodetic, UVW,
     };
+    use ndarray::{s, Array2, Array3};
     use std::f64::consts::PI;
 
     use crate::{
-        approx::assert_abs_diff_eq, context_to_jones_array, get_flaggable_timesteps, TestJones,
+        approx::assert_abs_diff_eq,
+        context_to_jones_array,
+        corrections::_correct_digital_gains,
+        flags::{get_coarse_chan_range, get_timestep_range},
+        get_flaggable_timesteps, BirliError, TestJones,
     };
 
     // TODO: Why does clippy think CxxImageSet.ImageBuffer() is &[f64]?
@@ -349,22 +496,22 @@ mod tests {
     fn test_cable_length_corrections_mwax() {
         let context = get_mwax_context();
 
-        let img_timestep_idxs = get_flaggable_timesteps(&context).unwrap();
-        assert_eq!(img_timestep_idxs.len(), 4);
-        let img_timestep_range =
-            *img_timestep_idxs.first().unwrap()..(*img_timestep_idxs.last().unwrap() + 1);
-        let img_coarse_chan_idxs = &context.common_coarse_chan_indices;
-        let img_coarse_chan_range =
-            *img_coarse_chan_idxs.first().unwrap()..(*img_coarse_chan_idxs.last().unwrap() + 1);
+        let sel_timestep_idxs = get_flaggable_timesteps(&context).unwrap();
+        assert_eq!(sel_timestep_idxs.len(), 4);
+        let sel_timestep_range =
+            *sel_timestep_idxs.first().unwrap()..(*sel_timestep_idxs.last().unwrap() + 1);
+        let sel_coarse_chan_idxs = &context.common_coarse_chan_indices;
+        let sel_coarse_chan_range =
+            *sel_coarse_chan_idxs.first().unwrap()..(*sel_coarse_chan_idxs.last().unwrap() + 1);
 
-        let all_freqs_hz = context.get_fine_chan_freqs_hz_array(img_coarse_chan_idxs);
+        let all_freqs_hz = context.get_fine_chan_freqs_hz_array(sel_coarse_chan_idxs);
 
-        // let img_baseline_idxs: Vec<usize> = (0..context.metafits_context.num_baselines).collect();
+        // let sel_baseline_idxs: Vec<usize> = (0..context.metafits_context.num_baselines).collect();
 
         let (jones_array, _) = context_to_jones_array(
             &context,
-            &img_timestep_range,
-            &img_coarse_chan_range,
+            &sel_timestep_range,
+            &sel_coarse_chan_range,
             None,
             false,
         )
@@ -469,7 +616,7 @@ mod tests {
 
         // TODO: this is not great.
         let mut jones_array = jones_array.mapv(|e| Jones::from([e[0], e[1], e[2], e[3]]));
-        correct_cable_lengths(&context, &mut jones_array, &img_coarse_chan_range, false);
+        correct_cable_lengths(&context, &mut jones_array, &sel_coarse_chan_range, false);
 
         let jones_array = jones_array.mapv(TestJones::from);
 
@@ -533,22 +680,22 @@ mod tests {
     fn test_cable_length_corrections_ord() {
         let context = get_mwa_ord_context();
 
-        let img_timestep_idxs = get_flaggable_timesteps(&context).unwrap();
-        assert_eq!(img_timestep_idxs.len(), 4);
-        let img_timestep_range =
-            *img_timestep_idxs.first().unwrap()..(*img_timestep_idxs.last().unwrap() + 1);
-        let img_coarse_chan_idxs = &context.common_coarse_chan_indices;
-        let img_coarse_chan_range =
-            *img_coarse_chan_idxs.first().unwrap()..(*img_coarse_chan_idxs.last().unwrap() + 1);
+        let sel_timestep_idxs = get_flaggable_timesteps(&context).unwrap();
+        assert_eq!(sel_timestep_idxs.len(), 4);
+        let sel_timestep_range =
+            *sel_timestep_idxs.first().unwrap()..(*sel_timestep_idxs.last().unwrap() + 1);
+        let sel_coarse_chan_idxs = &context.common_coarse_chan_indices;
+        let sel_coarse_chan_range =
+            *sel_coarse_chan_idxs.first().unwrap()..(*sel_coarse_chan_idxs.last().unwrap() + 1);
 
-        let all_freqs_hz = context.get_fine_chan_freqs_hz_array(img_coarse_chan_idxs);
+        let all_freqs_hz = context.get_fine_chan_freqs_hz_array(sel_coarse_chan_idxs);
 
-        // let img_baseline_idxs: Vec<usize> = (0..context.metafits_context.num_baselines).collect();
+        // let sel_baseline_idxs: Vec<usize> = (0..context.metafits_context.num_baselines).collect();
 
         let (jones_array, _) = context_to_jones_array(
             &context,
-            &img_timestep_range,
-            &img_coarse_chan_range,
+            &sel_timestep_range,
+            &sel_coarse_chan_range,
             None,
             false,
         )
@@ -654,7 +801,7 @@ mod tests {
         // FIXME: aaaaaaa
         let mut jones_array = jones_array.mapv(|e| Jones::from([e[0], e[1], e[2], e[3]]));
 
-        correct_cable_lengths(&context, &mut jones_array, &img_coarse_chan_range, false);
+        correct_cable_lengths(&context, &mut jones_array, &sel_coarse_chan_range, false);
 
         let jones_array = jones_array.mapv(TestJones::from);
 
@@ -718,15 +865,15 @@ mod tests {
     fn test_geometric_corrections_ord() {
         let context = get_mwa_ord_context();
 
-        let img_timestep_idxs = get_flaggable_timesteps(&context).unwrap();
-        assert_eq!(img_timestep_idxs.len(), 4);
-        let img_timestep_range =
-            *img_timestep_idxs.first().unwrap()..(*img_timestep_idxs.last().unwrap() + 1);
-        let img_coarse_chan_idxs = &context.common_coarse_chan_indices;
-        let img_coarse_chan_range =
-            *img_coarse_chan_idxs.first().unwrap()..(*img_coarse_chan_idxs.last().unwrap() + 1);
+        let sel_timestep_idxs = get_flaggable_timesteps(&context).unwrap();
+        assert_eq!(sel_timestep_idxs.len(), 4);
+        let sel_timestep_range =
+            *sel_timestep_idxs.first().unwrap()..(*sel_timestep_idxs.last().unwrap() + 1);
+        let sel_coarse_chan_idxs = &context.common_coarse_chan_indices;
+        let sel_coarse_chan_range =
+            *sel_coarse_chan_idxs.first().unwrap()..(*sel_coarse_chan_idxs.last().unwrap() + 1);
 
-        let all_freqs_hz = context.get_fine_chan_freqs_hz_array(img_coarse_chan_idxs);
+        let all_freqs_hz = context.get_fine_chan_freqs_hz_array(sel_coarse_chan_idxs);
 
         let array_pos = LatLngHeight::new_mwa();
 
@@ -737,8 +884,8 @@ mod tests {
 
         let (jones_array, _) = context_to_jones_array(
             &context,
-            &img_timestep_range,
-            &img_coarse_chan_range,
+            &sel_timestep_range,
+            &sel_coarse_chan_range,
             None,
             false,
         )
@@ -785,9 +932,10 @@ mod tests {
         let integration_time_s = context.metafits_context.corr_int_time_ms as f64 / 1000.0;
 
         // timestep 0
-        let timestep_0 = &context.timesteps[img_timestep_idxs[0]];
-        let epoch_0 =
-            time::gps_to_epoch(timestep_0.gps_time_ms as f64 / 1000.0 + integration_time_s / 2.0);
+        let timestep_0 = &context.timesteps[sel_timestep_idxs[0]];
+        let epoch_0 = Epoch::from_gpst_seconds(
+            timestep_0.gps_time_ms as f64 / 1000.0 + integration_time_s / 2.0,
+        );
         let prec_info_0 = precess_time(
             phase_centre_ra,
             epoch_0,
@@ -797,9 +945,10 @@ mod tests {
         let phase_centre_ha_j2000_0 = prec_info_0.hadec_j2000; // phase_centre_ra.to_hadec(prec_info_0.lmst_j2000);
         let tiles_xyz_precessed_0 = prec_info_0.precess_xyz_parallel(&tiles_xyz_geod);
         // timestep 3
-        let timestep_3 = &context.timesteps[img_timestep_idxs[3]];
-        let epoch_3 =
-            time::gps_to_epoch(timestep_3.gps_time_ms as f64 / 1000.0 + integration_time_s / 2.0);
+        let timestep_3 = &context.timesteps[sel_timestep_idxs[3]];
+        let epoch_3 = Epoch::from_gpst_seconds(
+            timestep_3.gps_time_ms as f64 / 1000.0 + integration_time_s / 2.0,
+        );
         let prec_info_3 = precess_time(
             phase_centre_ra,
             epoch_3,
@@ -838,8 +987,8 @@ mod tests {
         correct_geometry(
             &context,
             &mut jones_array,
-            &img_timestep_range,
-            &img_coarse_chan_range,
+            &sel_timestep_range,
+            &sel_coarse_chan_range,
             None,
             None,
             false,
@@ -907,14 +1056,14 @@ mod tests {
     fn test_geometric_corrections_mwax() {
         let context = get_mwax_context();
 
-        let img_timestep_idxs = get_flaggable_timesteps(&context).unwrap();
-        let img_timestep_range =
-            *img_timestep_idxs.first().unwrap()..(*img_timestep_idxs.last().unwrap() + 1);
-        let img_coarse_chan_idxs = &context.common_coarse_chan_indices;
-        let img_coarse_chan_range =
-            *img_coarse_chan_idxs.first().unwrap()..(*img_coarse_chan_idxs.last().unwrap() + 1);
+        let sel_timestep_idxs = get_flaggable_timesteps(&context).unwrap();
+        let sel_timestep_range =
+            *sel_timestep_idxs.first().unwrap()..(*sel_timestep_idxs.last().unwrap() + 1);
+        let sel_coarse_chan_idxs = &context.common_coarse_chan_indices;
+        let sel_coarse_chan_range =
+            *sel_coarse_chan_idxs.first().unwrap()..(*sel_coarse_chan_idxs.last().unwrap() + 1);
 
-        let all_freqs_hz = context.get_fine_chan_freqs_hz_array(img_coarse_chan_idxs);
+        let all_freqs_hz = context.get_fine_chan_freqs_hz_array(sel_coarse_chan_idxs);
 
         let array_pos = LatLngHeight::new_mwa();
 
@@ -925,9 +1074,9 @@ mod tests {
 
         let (jones_array, _) = context_to_jones_array(
             &context,
-            &img_timestep_range,
-            &img_coarse_chan_range,
-            // img_baseline_idxs.as_slice(),
+            &sel_timestep_range,
+            &sel_coarse_chan_range,
+            // sel_baseline_idxs.as_slice(),
             None,
             false,
         )
@@ -974,9 +1123,10 @@ mod tests {
         let integration_time_s = context.metafits_context.corr_int_time_ms as f64 / 1000.0;
 
         // timestep 0
-        let timestep_0 = &context.timesteps[img_timestep_idxs[0]];
-        let epoch_0 =
-            time::gps_to_epoch(timestep_0.gps_time_ms as f64 / 1000.0 + integration_time_s / 2.0);
+        let timestep_0 = &context.timesteps[sel_timestep_idxs[0]];
+        let epoch_0 = Epoch::from_gpst_seconds(
+            timestep_0.gps_time_ms as f64 / 1000.0 + integration_time_s / 2.0,
+        );
         let prec_info_0 = precess_time(
             phase_centre_ra,
             epoch_0,
@@ -986,9 +1136,10 @@ mod tests {
         let phase_centre_ha_j2000_0 = prec_info_0.hadec_j2000; // phase_centre_ra.to_hadec(prec_info_0.lmst_j2000);
         let tiles_xyz_precessed_0 = prec_info_0.precess_xyz_parallel(&tiles_xyz_geod);
         // timestep 3
-        let timestep_3 = &context.timesteps[img_timestep_idxs[3]];
-        let epoch_3 =
-            time::gps_to_epoch(timestep_3.gps_time_ms as f64 / 1000.0 + integration_time_s / 2.0);
+        let timestep_3 = &context.timesteps[sel_timestep_idxs[3]];
+        let epoch_3 = Epoch::from_gpst_seconds(
+            timestep_3.gps_time_ms as f64 / 1000.0 + integration_time_s / 2.0,
+        );
         let prec_info_3 = precess_time(
             phase_centre_ra,
             epoch_3,
@@ -1025,8 +1176,8 @@ mod tests {
         correct_geometry(
             &context,
             &mut jones_array,
-            &img_timestep_range,
-            &img_coarse_chan_range,
+            &sel_timestep_range,
+            &sel_coarse_chan_range,
             None,
             None,
             false,
@@ -1087,4 +1238,144 @@ mod tests {
             ])
         );
     }
+
+    macro_rules! compare_jones {
+        ($a:expr, $b:expr) => {
+            assert_abs_diff_eq!(TestJones::<f32>::from($a), TestJones::<f32>::from($b));
+        };
+    }
+
+    #[test]
+    fn test_correct_digital_gains() {
+        let context = get_mwa_ord_context();
+
+        let sel_coarse_chan_range = get_coarse_chan_range(&context).unwrap();
+        let sel_timestep_range = get_timestep_range(&context).unwrap();
+
+        let (jones_array, _) = context_to_jones_array(
+            &context,
+            &sel_timestep_range,
+            &sel_coarse_chan_range,
+            // sel_baseline_idxs.as_slice(),
+            None,
+            false,
+        )
+        .unwrap();
+
+        let sel_baseline_range = 0..2;
+        let sel_baseline_idxs: Vec<usize> = sel_baseline_range.clone().collect();
+
+        let sel_baselines = sel_baseline_idxs
+            .iter()
+            .map(|&idx| {
+                let baseline = &context.metafits_context.baselines[idx];
+                (baseline.ant1_index, baseline.ant2_index)
+            })
+            .collect::<Vec<_>>();
+
+        let first_coarse_chan_idx = sel_coarse_chan_range.start;
+        let last_coarse_chan_idx = sel_coarse_chan_range.end - 1;
+        // first coarse chan, antenna 0
+        let gain_first_0 =
+            context.metafits_context.antennas[0].rfinput_x.digital_gains[first_coarse_chan_idx];
+        // first coarse chan, antenna 1
+        let gain_first_1 =
+            context.metafits_context.antennas[1].rfinput_x.digital_gains[first_coarse_chan_idx];
+        // last coarse chan, antenna 0
+        let gain_last_0 =
+            context.metafits_context.antennas[0].rfinput_x.digital_gains[last_coarse_chan_idx];
+        // last coarse chan, antenna 1
+        let gain_last_1 =
+            context.metafits_context.antennas[1].rfinput_x.digital_gains[last_coarse_chan_idx];
+
+        let max_chan = jones_array.dim().1 - 1;
+
+        // ts 0, first chan, baseline 0 (0 vs 0)
+        let jones_0_min_0 = jones_array[(0, 0, 0)];
+        // ts 0, first chan, baseline 1 (0 vs 1)
+        let jones_0_min_1 = jones_array[(0, 0, 1)];
+        // ts 0, last chan, baseline 0 (0 vs 0)
+        let jones_0_max_0 = jones_array[(0, max_chan, 0)];
+        // ts 0, last chan, baseline 1 (0 vs 1)
+        let jones_0_max_1 = jones_array[(0, max_chan, 1)];
+
+        let mut out_jones_array = jones_array.slice(s![.., .., sel_baseline_range]).to_owned();
+
+        correct_digital_gains(
+            &context,
+            &mut out_jones_array,
+            &sel_coarse_chan_range,
+            &sel_baselines,
+        )
+        .unwrap();
+
+        compare_jones!(
+            out_jones_array[(0, 0, 0)],
+            jones_0_min_0 / ((gain_first_0 * gain_first_0) as f32)
+        );
+        compare_jones!(
+            out_jones_array[(0, 0, 1)],
+            jones_0_min_1 / ((gain_first_0 * gain_first_1) as f32)
+        );
+        compare_jones!(
+            out_jones_array[(0, max_chan, 0)],
+            jones_0_max_0 / ((gain_last_0 * gain_last_0) as f32)
+        );
+        compare_jones!(
+            out_jones_array[(0, max_chan, 1)],
+            jones_0_max_1 / ((gain_last_0 * gain_last_1) as f32)
+        );
+    }
+
+    #[test]
+    fn test_correct_digital_gains_bad_array_shape() {
+        let context = get_mwa_ord_context();
+
+        let sel_coarse_chan_range = get_coarse_chan_range(&context).unwrap();
+
+        let mut jones_array = Array3::from_shape_fn((1, 1, 1), |_| Jones::nan());
+
+        let sel_baseline_range = 0..2;
+        let sel_baseline_idxs: Vec<usize> = sel_baseline_range.collect();
+
+        let sel_baselines = sel_baseline_idxs
+            .iter()
+            .map(|&idx| {
+                let baseline = &context.metafits_context.baselines[idx];
+                (baseline.ant1_index, baseline.ant2_index)
+            })
+            .collect::<Vec<_>>();
+
+        assert!(matches!(
+            correct_digital_gains(
+                &context,
+                &mut jones_array,
+                &sel_coarse_chan_range,
+                &sel_baselines,
+            ),
+            Err(BirliError::BadArrayShape { .. })
+        ));
+
+        let channels =
+            sel_coarse_chan_range.len() * context.metafits_context.num_corr_fine_chans_per_coarse;
+
+        let mut jones_array = Array3::from_shape_fn((1, channels, 1), |_| Jones::nan());
+
+        assert!(matches!(
+            correct_digital_gains(
+                &context,
+                &mut jones_array,
+                &sel_coarse_chan_range,
+                &sel_baselines,
+            ),
+            Err(BirliError::BadArrayShape { .. })
+        ));
+
+        let gains = Array2::from_shape_fn((1, 1), |_| (1., 1.));
+
+        assert!(matches!(
+            _correct_digital_gains(&mut jones_array, &gains, &sel_baselines, 2),
+            Err(BirliError::BadArrayShape { .. })
+        ));
+    }
 }
